/*-
 * Copyright (c) 2007, 2008 Bartosz Fabianowski <freebsd@chillt.de>
 * All rights reserved.
 *
 * Financed by the "Irish Research Council for Science, Engineering and
 * Technology: funded by the National Development Plan"
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions, and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
 * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*-
 * Copyright (c) 1998 The NetBSD Foundation, Inc.
 * All rights reserved.
 *
 * This code is derived from software contributed to The NetBSD Foundation
 * by Lennart Augustsson (lennart@augustsson.net) at
 * Carlstedt Research & Technology.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *        This product includes software developed by the NetBSD
 *        Foundation, Inc. and its contributors.
 * 4. Neither the name of The NetBSD Foundation nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE NETBSD FOUNDATION, INC. AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/* $Id: uwacom.c 38 2008-11-13 03:01:11Z undo $
In sync with:
  /linuxwacom-dev/src/2.6.19/wacom_sys.c 1.7
  /linuxwacom-dev/src/2.6.19/wacom_wac.c 1.12
*/

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/clist.h>
#include <sys/kernel.h>
#include <sys/module.h>
#include <sys/bus.h>
#include <sys/conf.h>
#include <sys/tty.h>
#include <sys/malloc.h>
#include <sys/signalvar.h>
#include <sys/endian.h>
#include <sys/proc.h>
#include <sys/fcntl.h>
#include <sys/uio.h>
#include <sys/filio.h>
#include <sys/poll.h>

#include <dev/usb/usb.h>
#include <dev/usb/usbdi.h>
#include <dev/usb/usbdi_util.h>
#include <dev/usb/usbhid.h>

// USB IDs
#define USB_VENDOR_WACOM                 0x056a

#define USB_PRODUCT_WACOM_GRAPHIRE       0x0010
#define USB_PRODUCT_WACOM_GRAPHIRE2_4X5  0x0011
#define USB_PRODUCT_WACOM_GRAPHIRE2_5X7  0x0012
#define USB_PRODUCT_WACOM_GRAPHIRE3_4X5  0x0013
#define USB_PRODUCT_WACOM_GRAPHIRE3_6X8  0x0014
#define USB_PRODUCT_WACOM_GRAPHIRE4_4X5  0x0015
#define USB_PRODUCT_WACOM_GRAPHIRE4_6X8  0x0016
#define USB_PRODUCT_WACOM_BAMBOO_FUN_4X5 0x0017
#define USB_PRODUCT_WACOM_BAMBOO_FUN_6X8 0x0018
#define USB_PRODUCT_WACOM_CINTIQ_21UX    0x003f
#define USB_PRODUCT_WACOM_VOLITO         0x0060
#define USB_PRODUCT_WACOM_PENSTATION2    0x0061
#define USB_PRODUCT_WACOM_VOLITO2_4X5    0x0062
#define USB_PRODUCT_WACOM_VOLITO2_2X3    0x6003
#define USB_PRODUCT_WACOM_PENPARTNER2    0x0064
#define USB_PRODUCT_WACOM_BAMBOO         0x0065
#define USB_PRODUCT_WACOM_BAMBOO_ONE     0x0069
#define USB_PRODUCT_WACOM_INTUOS3_4X5    0x00b0
#define USB_PRODUCT_WACOM_INTUOS3_6X8    0x00b1
#define USB_PRODUCT_WACOM_INTUOS3_9X12   0x00b2
#define USB_PRODUCT_WACOM_INTUOS3_12X12  0x00b3
#define USB_PRODUCT_WACOM_INTUOS3_12X19  0x00b4
#define USB_PRODUCT_WACOM_INTUOS3_6X11   0x00b5
#define USB_PRODUCT_WACOM_INTUOS3_4X6    0x00b7

// Protocols
struct uwacom_protocol_type {
  int packet_size;
  int distance_max;
};

enum uwacom_protocol {
  // Graphire family
  GRAPHIRE = 0,
  GRAPHIRE4,
  MYOFFICE,

  // Intuos family
  INTUOS3S,
  INTUOS3L,
  CINTIQ
};

static const struct uwacom_protocol_type uwacom_protocols[] = {
  { 8, 63},
  { 8, 63},
  { 9, 63},
  {10, 63},
  {10, 63},
  {10, 63}
};

#define USB_PACKET_SIZE_MAX              10

// Models
struct uwacom_type {
  struct usb_devno     devno;
  enum uwacom_protocol protocol;
  int                  x_max;
  int                  y_max;
  int                  pressure_max;
};

static const struct uwacom_type uwacom_devs[] = {
  {{USB_VENDOR_WACOM, USB_PRODUCT_WACOM_GRAPHIRE},       GRAPHIRE,  10206,  7422,  511},
  {{USB_VENDOR_WACOM, USB_PRODUCT_WACOM_GRAPHIRE2_4X5},  GRAPHIRE,  10206,  7422,  511},
  {{USB_VENDOR_WACOM, USB_PRODUCT_WACOM_GRAPHIRE2_5X7},  GRAPHIRE,  13918, 10206,  511},
  {{USB_VENDOR_WACOM, USB_PRODUCT_WACOM_GRAPHIRE3_4X5},  GRAPHIRE,  10208,  7424,  511},
  {{USB_VENDOR_WACOM, USB_PRODUCT_WACOM_GRAPHIRE3_6X8},  GRAPHIRE,  16704, 12064,  511},
  {{USB_VENDOR_WACOM, USB_PRODUCT_WACOM_GRAPHIRE4_4X5},  GRAPHIRE4, 10208,  7424,  511},
  {{USB_VENDOR_WACOM, USB_PRODUCT_WACOM_GRAPHIRE4_6X8},  GRAPHIRE4, 16704, 12064,  511},
  {{USB_VENDOR_WACOM, USB_PRODUCT_WACOM_BAMBOO_FUN_4X5}, MYOFFICE,  14760,  9225,  511},
  {{USB_VENDOR_WACOM, USB_PRODUCT_WACOM_BAMBOO_FUN_6X8}, MYOFFICE,  21648, 13530,  511},
  {{USB_VENDOR_WACOM, USB_PRODUCT_WACOM_CINTIQ_21UX},    CINTIQ,    87200, 65600, 1023},
  {{USB_VENDOR_WACOM, USB_PRODUCT_WACOM_VOLITO},         GRAPHIRE,   5104,  3712,  511},
  {{USB_VENDOR_WACOM, USB_PRODUCT_WACOM_PENSTATION2},    GRAPHIRE,   3250,  2320,  255},
  {{USB_VENDOR_WACOM, USB_PRODUCT_WACOM_VOLITO2_4X5},    GRAPHIRE,   5104,  3712,  511},
  {{USB_VENDOR_WACOM, USB_PRODUCT_WACOM_VOLITO2_2X3},    GRAPHIRE,   3248,  2320,  511},
  {{USB_VENDOR_WACOM, USB_PRODUCT_WACOM_PENPARTNER2},    GRAPHIRE,   3250,  2320,  511},
  {{USB_VENDOR_WACOM, USB_PRODUCT_WACOM_BAMBOO},         MYOFFICE,  14760,  9225,  511},
  {{USB_VENDOR_WACOM, USB_PRODUCT_WACOM_BAMBOO_ONE},     GRAPHIRE,   5104,  3712,  511},
  {{USB_VENDOR_WACOM, USB_PRODUCT_WACOM_INTUOS3_4X5},    INTUOS3S,  25400, 20320, 1023},
  {{USB_VENDOR_WACOM, USB_PRODUCT_WACOM_INTUOS3_6X8},    INTUOS3L,  40640, 30480, 1023},
  {{USB_VENDOR_WACOM, USB_PRODUCT_WACOM_INTUOS3_9X12},   INTUOS3L,  60960, 45720, 1023},
  {{USB_VENDOR_WACOM, USB_PRODUCT_WACOM_INTUOS3_12X12},  INTUOS3L,  60960, 60960, 1023},
  {{USB_VENDOR_WACOM, USB_PRODUCT_WACOM_INTUOS3_12X19},  INTUOS3L,  97536, 60960, 1023},
  {{USB_VENDOR_WACOM, USB_PRODUCT_WACOM_INTUOS3_6X11},   INTUOS3L,  54204, 31750, 1023},
  {{USB_VENDOR_WACOM, USB_PRODUCT_WACOM_INTUOS3_4X6},    INTUOS3S,  31496, 19685, 1023}
  };

#define uwacom_lookup(v, p) ((const struct uwacom_type*)usb_lookup(uwacom_devs, v, p))

#define MINOR(t, h)                      (unit2minor(((t) << 16) | (h)))
#define TABLET(x)                        ((dev2unit(x) >> 16) & 0xff)
#define HANDLER(x)                       (dev2unit(x) & 0xffff)
#define HANDLER_MAX                      0x10000

#define PACKET_BIT(b, s)                 ((packet[b] >> (s)) & 1)
#define PACKET_BITS(b, s, n)             ((((s) + (n) > 32 ? ((packet[(b) - 4]) << (32 - (s))) : 0) | \
                                           ((s) + (n) > 24 ? ((packet[(b) - 3]) << (24 - (s))) : 0) | \
                                           ((s) + (n) > 16 ? ((packet[(b) - 2]) << (16 - (s))) : 0) | \
                                           ((s) + (n) >  8 ? ((packet[(b) - 1]) << ( 8 - (s))) : 0) | \
                                                             ((packet[(b)    ]) >>       (s))) & \
                                           ((n) == 32 ? 0xffffffff : (1 << (n)) - 1))

#define EUWACOMGETVERSION                0x01
#define EUWACOMGETID                     0x02
#define EUWACOMGETBM                     0x20
#define EUWACOMGETABS                    0x40

struct id {
  uint16_t bus;
#define ID_BUS_USB 3
  uint16_t vendor;
  uint16_t product;
  uint16_t version;
};

#define EVT_SYN                          0x0000
#define EVT_BTN                          0x0001
#define EVT_REL                          0x0002
#define EVT_ABS                          0x0003
#define EVT_MSC                          0x0004
#define EVT_USED                         0x0005
#define EVT_MAX                          0x001f

#define SYN_REPORT                       0x0000

#define BTN_MISC_0                       0x0100
#define BTN_MISC_1                       0x0101
#define BTN_MISC_2                       0x0102
#define BTN_MISC_3                       0x0103
#define BTN_MISC_4                       0x0104
#define BTN_MISC_5                       0x0105
#define BTN_MISC_6                       0x0106
#define BTN_MISC_7                       0x0107
#define BTN_LEFT                         0x0110
#define BTN_RIGHT                        0x0111
#define BTN_MIDDLE                       0x0112
#define BTN_SIDE                         0x0113
#define BTN_EXTRA                        0x0114
#define BTN_TOOL_PEN                     0x0140
#define BTN_TOOL_ERASER                  0x0141
#define BTN_TOOL_PAD                     0x0145
#define BTN_TOOL_MOUSE                   0x0146
#define BTN_STYLUS_1                     0x014b
#define BTN_STYLUS_2                     0x014c
#define BTN_USED                         0x014d
#define BTN_MAX                          0x01ff

#define REL_WHEEL                        0x0008
#define REL_MAX                          0x000f

#define ABS_X                            0x0000
#define ABS_Y                            0x0001
#define ABS_Z                            0x0002
#define ABS_RX                           0x0003
#define ABS_RY                           0x0004
#define ABS_WHEEL                        0x0008
#define ABS_PRESSURE                     0x0018
#define ABS_DISTANCE                     0x0019
#define ABS_TILT_X                       0x001a
#define ABS_TILT_Y                       0x001b
#define ABS_MISC                         0x0028
#define ABS_USED                         0x0029
#define ABS_MAX                          0x003f

#define MSC_SERIAL                       0x0000
#define MSC_MAX                          0x0007

#define BM_SIZE(x)                       (((x) / (sizeof(long) * 8) + 1) * sizeof(long))
#define BM_SET_BIT(x, y)                 ((x)[(y) / (sizeof(long) * 8)] |= (1ul << ((y) % (sizeof(long) * 8))))

static const size_t bm_size[EVT_USED] = {
  BM_SIZE(EVT_MAX),
  BM_SIZE(BTN_MAX),
  BM_SIZE(REL_MAX),
  BM_SIZE(ABS_MAX),
  BM_SIZE(MSC_MAX),
};

struct abs_axis {
  int32_t value;
  int32_t min;
  int32_t max;
  int32_t fuzz;
  int32_t flat;
};

struct event {
  struct timeval time;
  uint16_t type;
  uint16_t index;
  int32_t value;
};

#define TOOL_ID_PEN                      0x0002
#define TOOL_ID_MOUSE                    0x0006
#define TOOL_ID_ERASER                   0x000a
#define TOOL_ID_PAD                      0x000f

#define SERIAL_PAD_INTUOS3               0xffffffff
#define SERIAL_PAD_GRAPHIRE4             0x000000f0

d_open_t  uwacomopen;
d_close_t uwacomclose;
d_read_t  uwacomread;
d_ioctl_t uwacomioctl;
d_poll_t  uwacompoll;
d_purge_t uwacompurge;

static struct cdevsw uwacomphy_cdevsw = {
  .d_version = D_VERSION,
  .d_flags   = D_NEEDGIANT,
  .d_name    = "uwacomphy",
};

static struct cdevsw uwacom_cdevsw = {
  .d_version = D_VERSION,
  .d_flags   = D_NEEDGIANT,
  .d_open    = uwacomopen,
  .d_close   = uwacomclose,
  .d_read    = uwacomread,
  .d_ioctl   = uwacomioctl,
  .d_poll    = uwacompoll,
  .d_purge   = uwacompurge,
  .d_name    = "uwacom",
};

struct uwacom_handler {
  LIST_ENTRY(uwacom_handler) link;

  struct clist               queue;
  struct sigio              *queue_sigio;
  struct selinfo             rsel;

  u_char                     sleeping;
  u_char                     async;
};

struct uwacom_softc {
  device_t                    sc_dev;
  struct cdev                *dev;

  usbd_interface_handle       sc_iface;
  int                         sc_ep_addr;
  usbd_pipe_handle            sc_intrpipe;
  u_char                      sc_buf[USB_PACKET_SIZE_MAX];

  u_char                      sc_stalled;
  u_char                      sc_dying;

  const struct uwacom_type   *sc_type;
  struct id                   sc_id;
  unsigned long              *sc_bm[EVT_USED];

  int                        *sc_btn;
  struct abs_axis            *sc_abs;
  int                         sc_tool[2];
  int                         sc_tool_id[2];
  unsigned int                sc_serial[2];
  int                         sc_sync;

  LIST_HEAD(, uwacom_handler) sc_handlers;
};

static eventhandler_tag uwacom_ehtag = NULL;
static int uwacom_ehtag_ref = 0;

static void uwacom_intr(usbd_xfer_handle, usbd_private_handle, usbd_status);
static void uwacom_process_packet(struct uwacom_softc*);
static void uwacom_process_packet_graphire(struct uwacom_softc*);
void uwacom_pos_events_graphire(struct uwacom_softc*, int, int);
void uwacom_pen_events_graphire(struct uwacom_softc*, int, int, int);
void uwacom_mouse_events_graphire(struct uwacom_softc*, int, int, int, int, int);
void uwacom_pad_events_graphire4(struct uwacom_softc*, int, int, int, int, int, int);
void uwacom_tool_events_graphire(struct uwacom_softc*, int, int);
static void uwacom_process_packet_intuos(struct uwacom_softc*);
void uwacom_pos_events_intuos(struct uwacom_softc*, int, int, int);
void uwacom_pen_events_intuos(struct uwacom_softc*, int, int, int, int, int);
void uwacom_mouse_events_intuos(struct uwacom_softc*, int, int, int, int, int, int);
void uwacom_pad_events_intuos(struct uwacom_softc*, int, int, int, int, int, int, int, int, int, int);
void uwacom_tool_events_intuos(struct uwacom_softc*, int, int);
static void uwacom_event(struct uwacom_softc*, unsigned int, unsigned int, int);
static int uwacom_init(struct uwacom_softc*, usb_device_descriptor_t*);
static void uwacom_init_graphire4(struct uwacom_softc*);
static void uwacom_init_myoffice(struct uwacom_softc*);
static void uwacom_init_intuos(struct uwacom_softc*);
static void uwacom_init_intuos3_large(struct uwacom_softc*);
static void uwacom_init_abs(struct uwacom_softc*, int, int32_t, int32_t, int32_t, int32_t);
static void uwacom_clone(void*, struct ucred*, char*, int, struct cdev**);

static device_probe_t uwacom_match;
static device_attach_t uwacom_attach;
static device_detach_t uwacom_detach;

static devclass_t uwacom_devclass;

static device_method_t uwacom_methods[] = {
  DEVMETHOD(device_probe, uwacom_match),
  DEVMETHOD(device_attach, uwacom_attach),
  DEVMETHOD(device_detach, uwacom_detach),
  {0,0}
};

static driver_t uwacom_driver = {
  "uwacom",
  uwacom_methods,
  sizeof(struct uwacom_softc)
};

static int uwacom_match(device_t self) {
  struct usb_attach_arg *uaa = device_get_ivars(self);

  if (!uaa->iface)
    return UMATCH_NONE;

  return (uwacom_lookup(uaa->vendor, uaa->product) ? UMATCH_VENDOR_PRODUCT : UMATCH_NONE);
}

static int uwacom_attach(device_t self) {
  struct uwacom_softc *sc = device_get_softc(self);
  struct usb_attach_arg *uaa = device_get_ivars(self);

  sc->sc_dev = self;
  sc->sc_type = uwacom_lookup(uaa->vendor, uaa->product);
  if (!sc->sc_type) {
    printf("%s: unsupported tablet model\n", device_get_nameunit(sc->sc_dev));
    sc->sc_dying = 1;
    return ENXIO;
  }

  usb_device_descriptor_t *dd = usbd_get_device_descriptor(uaa->device);
  if (!dd) {
    printf("%s: could not read device descriptor\n", device_get_nameunit(sc->sc_dev));
    sc->sc_dying = 1;
    return ENXIO;
  }

  usb_endpoint_descriptor_t *ed = usbd_interface2endpoint_descriptor(uaa->iface, 0);
  if (!ed) {
    printf("%s: could not read endpoint descriptor\n", device_get_nameunit(sc->sc_dev));
    sc->sc_dying = 1;
    return ENXIO;
  }

  if (UE_GET_DIR(ed->bEndpointAddress) != UE_DIR_IN || (ed->bmAttributes & UE_XFERTYPE) != UE_INTERRUPT) {
    printf("%s: unexpected endpoint\n", device_get_nameunit(sc->sc_dev));
    sc->sc_dying = 1;
    return ENXIO;
  }

  if (uwacom_init(sc, dd))
    return ENXIO;

  static u_char report[] = {2, 2};
  for (int i = 0; i < 5; ++i)
    usbd_set_report(uaa->iface, UHID_FEATURE_REPORT, 2, &report, sizeof report);

  if (!uwacom_ehtag_ref++)
    uwacom_ehtag = EVENTHANDLER_REGISTER(dev_clone, uwacom_clone, 0, 1000);

  sc->dev        = make_dev(&uwacomphy_cdevsw, MINOR(device_get_unit(self), 0), UID_ROOT, GID_OPERATOR, 0644, "uwacom%d", device_get_unit(self));
  sc->sc_iface   = uaa->iface;
  sc->sc_ep_addr = ed->bEndpointAddress;

  return 0;
}

static int uwacom_detach(device_t self) {
  struct uwacom_softc *sc = device_get_softc(self);

  if (--uwacom_ehtag_ref == 0)
    EVENTHANDLER_DEREGISTER(dev_clone, uwacom_ehtag);

  sc->sc_dying = 1;

  if (sc->sc_intrpipe)
    usbd_abort_pipe(sc->sc_intrpipe);

  for (int i = 0; i < EVT_USED; ++i)
    free(sc->sc_bm[i], M_USBDEV);
  free(sc->sc_btn, M_USBDEV);
  free(sc->sc_abs, M_USBDEV);

  destroy_dev(sc->dev);

  return 0;
}

int uwacomopen(struct cdev *dev, int flag, int mode, struct thread *p) {
  struct uwacom_softc *sc = devclass_get_softc(uwacom_devclass, TABLET(dev));

  if (!sc)
    return ENXIO;

  if (sc->sc_dying)
    return ENXIO;

  if (dev->si_drv1)
    return EBUSY;

  struct uwacom_handler *handler = malloc(sizeof(struct uwacom_handler), M_USBDEV, M_NOWAIT | M_ZERO);
  if (!handler)
    return ENOMEM;

  clist_alloc_cblocks(&handler->queue, 0x10000, 0x10000);

  if ((!sc->sc_intrpipe) &&
      usbd_open_pipe_intr(sc->sc_iface, sc->sc_ep_addr, USBD_SHORT_XFER_OK, &sc->sc_intrpipe, sc,
          sc->sc_buf, uwacom_protocols[sc->sc_type->protocol].packet_size, uwacom_intr, USBD_DEFAULT_INTERVAL)) {
    clist_free_cblocks(&handler->queue);
    free(handler, M_USBDEV);
    return EIO;
  }

  dev->si_drv1 = handler;

  int s = splusb();
  LIST_INSERT_HEAD(&sc->sc_handlers, handler, link);
  splx(s);

  return 0;
}

int uwacomclose(struct cdev *dev, int flag, int mode, struct thread *p) {
  struct uwacom_softc *sc = devclass_get_softc(uwacom_devclass, TABLET(dev));

  if (!sc)
    return ENXIO;

  struct uwacom_handler *handler = (struct uwacom_handler*)dev->si_drv1;

  int s = splusb();
  LIST_REMOVE(handler, link);
  splx(s);

  ndflush(&handler->queue, handler->queue.c_cc);
  clist_free_cblocks(&handler->queue);

  funsetown(&handler->queue_sigio);

  free(handler, M_USBDEV);

  dev->si_drv1 = NULL;

  if (LIST_EMPTY(&sc->sc_handlers) && sc->sc_intrpipe) {
    usbd_abort_pipe(sc->sc_intrpipe);
    usbd_close_pipe(sc->sc_intrpipe);
    sc->sc_intrpipe = 0;
    sc->sc_stalled = 0;
  }

  return 0;
}

int uwacomread(struct cdev *dev, struct uio *uio, int flag) {
  struct uwacom_softc *sc = devclass_get_softc(uwacom_devclass, TABLET(dev));
  struct uwacom_handler *handler = (struct uwacom_handler*)dev->si_drv1;

  int error = 0;

  int s = splusb();
  while (!handler->queue.c_cc) {
    if (flag & O_NONBLOCK) {
      splx(s);
      return EWOULDBLOCK;
    }

    handler->sleeping = 1;
    error = tsleep(&handler->queue, PZERO | PCATCH, "uwacom", 0);
    handler->sleeping = 0;

    if (sc->sc_dying)
      error = EIO;

    if (error)
      break;

    if (sc->sc_stalled) {
      sc->sc_stalled = 0;
      usbd_clear_endpoint_stall(sc->sc_intrpipe);
    }
  }

  u_char buffer[128];
  while (handler->queue.c_cc > 0 && uio->uio_resid > 0 && !error) {
    size_t length = min(min(handler->queue.c_cc, uio->uio_resid), sizeof(buffer));

    q_to_b(&handler->queue, buffer, length);
    if ((error = uiomove(buffer, length, uio)) != 0)
      break;
  }
  splx(s);

  return error;
}

int uwacomioctl(struct cdev *dev, u_long cmd, caddr_t addr, int flag, struct thread *p) {
  struct uwacom_softc *sc = devclass_get_softc(uwacom_devclass, TABLET(dev));

  if (sc->sc_dying)
    return EIO;

  struct uwacom_handler *handler = (struct uwacom_handler*)dev->si_drv1;

  switch (cmd) {
    case FIONREAD: {
      int s = splusb();
      *(int*)addr = handler->queue.c_cc;
      splx(s);
      break;
    }

    case FIONBIO:
      break;

    case FIOASYNC:
      handler->async = *(int*)addr;
      break;

    case FIOSETOWN:
      return fsetown(*(int*)addr, &handler->queue_sigio);

    default: {
      int num = cmd & 0xff;
      int len = IOCPARM_LEN(cmd);

      if (IOCGROUP(cmd) != 'E' || (cmd & IOC_DIRMASK) != IOC_OUT)
        return EINVAL;

      switch (num) {
        case EUWACOMGETVERSION:
          if (len != sizeof(int))
            return EINVAL;
          *(int*)addr = 0x00010000;
          break;

        case EUWACOMGETID:
          if (len != sizeof(struct id))
            return EINVAL;
          memcpy(addr, &sc->sc_id, sizeof(struct id));
          break;

        default:
          if (num >= EUWACOMGETBM && num < EUWACOMGETBM + EVT_USED) {
            int idx = num - EUWACOMGETBM;
            size_t length = min(bm_size[idx], len);
            memcpy(addr, sc->sc_bm[idx], length);
            // Returning a positive value from an ioctl is somewhat of a gamble
            // Sometimes it has to be in td_retval[0], sometimes in td_retval[1]
            p->td_retval[0] = length;
            p->td_retval[1] = length;
            return 0;

          } else if (num >= EUWACOMGETABS && num < EUWACOMGETABS + ABS_USED) {
            if (len != sizeof(struct abs_axis))
              return EINVAL;
            int idx = num - EUWACOMGETABS;
            memcpy(addr, &sc->sc_abs[idx], sizeof(struct abs_axis));

          } else
            return EINVAL;
      }
    }
  }

  return 0;
}

int uwacompoll(struct cdev *dev, int events, struct thread *p) {
  struct uwacom_softc *sc = devclass_get_softc(uwacom_devclass, TABLET(dev));
  struct uwacom_handler *handler = (struct uwacom_handler*)dev->si_drv1;

  if (sc->sc_dying)
    return EIO;

  int revents = 0;

  if (events & (POLLIN | POLLRDNORM)) {
    int s = splusb();
    if (handler->queue.c_cc)
      revents = events & (POLLIN | POLLRDNORM);
    else
      selrecord(p, &handler->rsel);
    splx(s);
  }

  return revents;
}

void uwacompurge(struct cdev *dev) {
  struct uwacom_handler *handler = (struct uwacom_handler*)dev->si_drv1;

  if (handler->sleeping)
    wakeup(&handler->queue);
  selwakeuppri(&handler->rsel, PZERO);
}

static void uwacom_intr(usbd_xfer_handle xfer, usbd_private_handle addr, usbd_status status) {
  struct uwacom_softc *sc = addr;

  if (status == USBD_CANCELLED)
    return;

  if (status != USBD_NORMAL_COMPLETION) {
    if (status == USBD_STALLED)
      sc->sc_stalled = 1;

    return;
  }

  uwacom_process_packet(sc);

  struct uwacom_handler *handler;
  LIST_FOREACH(handler, &sc->sc_handlers, link)
    if (handler->queue.c_cc) {
      if (handler->sleeping) {
        handler->sleeping = 0;
        wakeup(&handler->queue);
      }
      selwakeuppri(&handler->rsel, PZERO);
      if (handler->async && handler->queue_sigio)
        pgsigio(&handler->queue_sigio, SIGIO, 0);
    }
}

static void uwacom_process_packet(struct uwacom_softc *sc) {
  switch (sc->sc_type->protocol) {
    case GRAPHIRE:
    case GRAPHIRE4:
    case MYOFFICE:
      uwacom_process_packet_graphire(sc);
      break;

    case INTUOS3S:
    case INTUOS3L:
    case CINTIQ:
      uwacom_process_packet_intuos(sc);
  }
}

static void uwacom_process_packet_graphire(struct uwacom_softc *sc) {
  u_char *packet = sc->sc_buf;

  if (PACKET_BITS(0, 0, 8) != 0x02) {
    printf("%s: unknown report type %02x received\n", device_get_nameunit(sc->sc_dev), PACKET_BITS(0, 0, 8));
    return;
  }

  // Tool in proximity
  if (PACKET_BIT(1, 7)) {
    uwacom_pos_events_graphire(sc, (PACKET_BITS(3, 0, 8) << 8) | PACKET_BITS(2, 0, 8), (PACKET_BITS(5, 0, 8) << 8) | PACKET_BITS(4, 0, 8));

    if (!PACKET_BIT(1, 6)) {
      if (!PACKET_BIT(1, 5)) {
        sc->sc_tool[0] = BTN_TOOL_PEN;
        sc->sc_tool_id[0] = TOOL_ID_PEN;
      } else {
        sc->sc_tool[0] = BTN_TOOL_ERASER;
        sc->sc_tool_id[0] = TOOL_ID_ERASER;
      }

      uwacom_pen_events_graphire(sc, (PACKET_BIT(7, 0) << 8) | PACKET_BITS(6, 0, 8), PACKET_BIT(1, 1), PACKET_BIT(1, 2));
    } else {
      int wheel, distance;

      if (sc->sc_type->protocol == GRAPHIRE) {
        wheel = (PACKET_BIT(1, 5) ? 0 : -(int8_t)PACKET_BITS(6, 0, 8));
        distance = PACKET_BITS(7, 0, 6);
      } else {
        wheel = (PACKET_BIT(7, 2) << 2) - PACKET_BITS(7, 0, 2);
        distance = PACKET_BITS(6, 0, 6);
      }

      sc->sc_tool[0] = BTN_TOOL_MOUSE;
      sc->sc_tool_id[0] = TOOL_ID_MOUSE;

      uwacom_mouse_events_graphire(sc, PACKET_BIT(1, 0), PACKET_BIT(1, 2), PACKET_BIT(1, 1), wheel, distance);
    }

    uwacom_tool_events_graphire(sc, 0, 1);

  // Tool leaving proximity
  } else if (sc->sc_tool_id[0]) {
    uwacom_pos_events_graphire(sc, 0, 0);

    if (sc->sc_tool[0] == BTN_TOOL_MOUSE)
      uwacom_mouse_events_graphire(sc, 0, 0, 0, 0, 0);
    else
      uwacom_pen_events_graphire(sc, 0, 0, 0);

    sc->sc_tool_id[0] = 0;
    uwacom_tool_events_graphire(sc, 0, 0);
  }

  // Finger on pad: Graphire4
  if ((sc->sc_type->protocol == GRAPHIRE4) && PACKET_BITS(7, 3, 5)) {
      sc->sc_tool_id[1] = TOOL_ID_PAD;
      uwacom_pad_events_graphire4(sc, PACKET_BIT(7, 6), 0, PACKET_BIT(7, 7), 0, PACKET_BITS(7, 3, 2) - (PACKET_BIT(7, 5) << 2), 0);

  // Finger on pad: MyOffice
  } else if ((sc->sc_type->protocol == MYOFFICE) && (PACKET_BITS(7, 3, 4) || PACKET_BITS(8, 0, 8))) {
      sc->sc_tool_id[1] = TOOL_ID_PAD;
      uwacom_pad_events_graphire4(sc, PACKET_BIT(7, 3), PACKET_BIT(7, 4), PACKET_BIT(7, 5), PACKET_BIT(7, 6), 0, PACKET_BITS(8, 0, 7));

  // Finger leaving pad
  } else if (sc->sc_tool_id[1]) {
    sc->sc_tool_id[1] = 0;
    uwacom_pad_events_graphire4(sc, 0, 0, 0, 0, 0, 0);
  }
}

void uwacom_pos_events_graphire(struct uwacom_softc *sc, int x, int y) {
  uwacom_event(sc, EVT_ABS, ABS_X, x);
  uwacom_event(sc, EVT_ABS, ABS_Y, y);
}

void uwacom_pen_events_graphire(struct uwacom_softc *sc, int pressure, int stylus_1, int stylus_2) {
  uwacom_event(sc, EVT_ABS, ABS_PRESSURE, pressure);
  uwacom_event(sc, EVT_BTN, BTN_STYLUS_1, stylus_1);
  uwacom_event(sc, EVT_BTN, BTN_STYLUS_2, stylus_2);
}

void uwacom_mouse_events_graphire(struct uwacom_softc *sc, int left, int middle, int right, int wheel, int distance) {
  uwacom_event(sc, EVT_BTN, BTN_LEFT, left);
  uwacom_event(sc, EVT_BTN, BTN_MIDDLE, middle);
  uwacom_event(sc, EVT_BTN, BTN_RIGHT, right);
  uwacom_event(sc, EVT_REL, REL_WHEEL, wheel);
  uwacom_event(sc, EVT_ABS, ABS_DISTANCE, distance);
}

void uwacom_pad_events_graphire4(struct uwacom_softc *sc, int btn_0, int btn_1, int btn_4, int btn_5, int rel, int abs) {
  uwacom_event(sc, EVT_BTN, BTN_MISC_0, btn_0);
  uwacom_event(sc, EVT_BTN, BTN_MISC_1, btn_1);
  uwacom_event(sc, EVT_BTN, BTN_MISC_4, btn_4);
  uwacom_event(sc, EVT_BTN, BTN_MISC_5, btn_5);
  uwacom_event(sc, EVT_REL, REL_WHEEL, rel);
  uwacom_event(sc, EVT_ABS, ABS_WHEEL, abs);
  uwacom_tool_events_graphire(sc, 1, btn_0 | btn_1 | btn_4 | btn_5 | rel | abs);
}

void uwacom_tool_events_graphire(struct uwacom_softc *sc, int idx, int proximity) {
  uwacom_event(sc, EVT_BTN, sc->sc_tool[idx], proximity);
  uwacom_event(sc, EVT_ABS, ABS_MISC, sc->sc_tool_id[idx]);
  if (sc->sc_serial[idx])
    uwacom_event(sc, EVT_MSC, MSC_SERIAL, sc->sc_serial[idx]);
  uwacom_event(sc, EVT_SYN, SYN_REPORT, 0);
}

static void uwacom_process_packet_intuos(struct uwacom_softc *sc) {
  u_char *packet = sc->sc_buf;

  switch (PACKET_BITS(0, 0, 8)) {
    case 0x02:
      switch (PACKET_BITS(1, 5, 2)) {
        // Tool entering proximity
        case 0x2:
          sc->sc_tool_id[0] = PACKET_BITS(3, 4, 12);
          sc->sc_serial[0] = (PACKET_BIT(1, 1) ? PACKET_BITS(7, 4, 32) : 0);

          switch (sc->sc_tool_id[0]) {
            case 0x823: // Intuos3 Grip Pen
            case 0x885: // Intuos3 Art Marker
              sc->sc_tool[0] = BTN_TOOL_PEN;
              break;
            case 0x82b: // Intuos3 Grip Pen eraser
              sc->sc_tool[0] = BTN_TOOL_ERASER;
              break;
            case 0x017: // Intuos3 2D mouse
              sc->sc_tool[0] = BTN_TOOL_MOUSE;
              break;
            default:
              printf("%s: unknown tool ID %03x seen\n", device_get_nameunit(sc->sc_dev), sc->sc_tool_id[0]);
              sc->sc_tool[0] = BTN_TOOL_PEN;
          }
          break;

        // Tool leaving proximity
        case 0x0:
          uwacom_pos_events_intuos(sc, 0, 0, 0);

          if (sc->sc_tool[0] == BTN_TOOL_MOUSE)
            uwacom_mouse_events_intuos(sc, 0, 0, 0, 0, 0, 0);
          else
            uwacom_pen_events_intuos(sc, 0, 0, 0, 0, 0);

          sc->sc_tool_id[0] = 0;
          uwacom_tool_events_intuos(sc, 0, 0);
          break;

        // Tool motion, outbound
        case 0x1:
          // Outbound tracking is unreliable on the Cintiq
          if (sc->sc_type->protocol == CINTIQ)
            break;

        // Tool motion
        case 0x3:
          uwacom_pos_events_intuos(sc, (PACKET_BITS(3, 0, 16) << 1) | PACKET_BIT(9, 1), (PACKET_BITS(5, 0, 16) << 1) | PACKET_BIT(9, 0), PACKET_BITS(9, 2, 6));

          if (PACKET_BITS(1, 3, 2) == 0)
            uwacom_pen_events_intuos(sc, PACKET_BITS(7, 6, 10), PACKET_BITS(8, 7, 7), PACKET_BITS(8, 0, 7), PACKET_BIT(1, 1), PACKET_BIT(1, 2));

          else if (PACKET_BITS(1, 1, 4) == 0x5) {
            int angle = 450 - PACKET_BITS(7, 6, 10);
            if (PACKET_BIT(7, 5))
              angle = (angle > 0 ? 900 : -900) - angle;
            uwacom_event(sc, EVT_ABS, ABS_Z, angle);
            break;

          } else if (PACKET_BITS(1, 1, 4) == 0x8)
            uwacom_mouse_events_intuos(sc, PACKET_BIT(8, 2), PACKET_BIT(8, 3), PACKET_BIT(8, 4), PACKET_BIT(8, 5), PACKET_BIT(8, 6),
                                           PACKET_BIT(8, 0) - PACKET_BIT(8, 1));

          else
            printf("%s: unsupported motion packet type %x received\n", device_get_nameunit(sc->sc_dev), PACKET_BITS(1, 1, 4));

          uwacom_tool_events_intuos(sc, 0, 1);
          break;
      }
      break;

    case 0x0c:
      uwacom_pad_events_intuos(sc, PACKET_BIT(5, 0), PACKET_BIT(5, 1), PACKET_BIT(5, 2), PACKET_BIT(5, 3),
                                   PACKET_BIT(6, 0), PACKET_BIT(6, 1), PACKET_BIT(6, 2), PACKET_BIT(6, 3),
                                   PACKET_BITS(2, 0, 13), PACKET_BITS(4, 0, 13));
      break;

    default:
      printf("%s: unknown report type %02x received\n", device_get_nameunit(sc->sc_dev), PACKET_BITS(0, 0, 8));
  }
}

void uwacom_pos_events_intuos(struct uwacom_softc *sc, int x, int y, int distance) {
  uwacom_event(sc, EVT_ABS, ABS_X, x);
  uwacom_event(sc, EVT_ABS, ABS_Y, y);
  uwacom_event(sc, EVT_ABS, ABS_DISTANCE, distance);
}

void uwacom_pen_events_intuos(struct uwacom_softc *sc, int pressure, int tilt_x, int tilt_y, int stylus_1, int stylus_2) {
  uwacom_event(sc, EVT_ABS, ABS_PRESSURE, pressure);
  uwacom_event(sc, EVT_ABS, ABS_TILT_X, tilt_x);
  uwacom_event(sc, EVT_ABS, ABS_TILT_Y, tilt_y);
  uwacom_event(sc, EVT_BTN, BTN_STYLUS_1, stylus_1);
  uwacom_event(sc, EVT_BTN, BTN_STYLUS_2, stylus_2);
}

void uwacom_mouse_events_intuos(struct uwacom_softc *sc, int left, int middle, int right, int extra, int side, int wheel) {
  uwacom_event(sc, EVT_BTN, BTN_LEFT, left);
  uwacom_event(sc, EVT_BTN, BTN_MIDDLE, middle);
  uwacom_event(sc, EVT_BTN, BTN_RIGHT, right);
  uwacom_event(sc, EVT_BTN, BTN_EXTRA, extra);
  uwacom_event(sc, EVT_BTN, BTN_SIDE, side);
  uwacom_event(sc, EVT_REL, REL_WHEEL, wheel);
}

void uwacom_pad_events_intuos(struct uwacom_softc *sc, int btn_0, int btn_1, int btn_2, int btn_3, int btn_4, int btn_5, int btn_6, int btn_7, int rx, int ry) {
  uwacom_event(sc, EVT_BTN, BTN_MISC_0, btn_0);
  uwacom_event(sc, EVT_BTN, BTN_MISC_1, btn_1);
  uwacom_event(sc, EVT_BTN, BTN_MISC_2, btn_2);
  uwacom_event(sc, EVT_BTN, BTN_MISC_3, btn_3);
  uwacom_event(sc, EVT_BTN, BTN_MISC_4, btn_4);
  uwacom_event(sc, EVT_BTN, BTN_MISC_5, btn_5);
  uwacom_event(sc, EVT_BTN, BTN_MISC_6, btn_6);
  uwacom_event(sc, EVT_BTN, BTN_MISC_7, btn_7);
  uwacom_event(sc, EVT_ABS, ABS_RX, rx);
  uwacom_event(sc, EVT_ABS, ABS_RY, ry);
  uwacom_tool_events_intuos(sc, 1, btn_0 | btn_1 | btn_2 | btn_3 | btn_4 | btn_5 | btn_6 | btn_7 | rx | ry);
}

void uwacom_tool_events_intuos(struct uwacom_softc *sc, int idx, int proximity) {
  uwacom_event(sc, EVT_BTN, sc->sc_tool[idx], proximity);
  uwacom_event(sc, EVT_ABS, ABS_MISC, sc->sc_tool_id[idx]);
  uwacom_event(sc, EVT_MSC, MSC_SERIAL, sc->sc_serial[idx]);
  uwacom_event(sc, EVT_SYN, SYN_REPORT, 0);
}

static void uwacom_event(struct uwacom_softc *sc, unsigned int type, unsigned int index, int value) {
  switch (type) {
    case EVT_SYN:
      if (sc->sc_sync)
        return;

      break;

    case EVT_BTN:
      if (sc->sc_btn[index] == value)
        return;

      sc->sc_btn[index] = value;
      break;

    case EVT_ABS:
      if (sc->sc_abs[index].fuzz) {
        int dist = abs(value - sc->sc_abs[index].value);

        if (dist < sc->sc_abs[index].fuzz >> 1)
          return;
        else if (dist < sc->sc_abs[index].fuzz)
          value = (7 * sc->sc_abs[index].value + value) >> 3;
        else if (dist < sc->sc_abs[index].fuzz << 1)
          value = (sc->sc_abs[index].value + value) >> 1;
      }

      if (sc->sc_abs[index].value == value)
        return;

      sc->sc_abs[index].value = value;
      break;

    case EVT_REL:
      if (!value)
        return;

      break;

    case EVT_MSC:
      break;

    default:
      return;
  }

  struct event ev;
  microtime(&ev.time);
  ev.type = type;
  ev.index = index;
  ev.value = value;

  struct uwacom_handler *handler;
  LIST_FOREACH(handler, &sc->sc_handlers, link)
    if (b_to_q((char*)&ev, sizeof(struct event), &handler->queue))
      printf("%s: event queue overflow\n", device_get_nameunit(sc->sc_dev));

  sc->sc_sync = (type == EVT_SYN);
}

static int uwacom_init(struct uwacom_softc *sc, usb_device_descriptor_t *dd) {
  sc->sc_id.bus     = ID_BUS_USB;
  sc->sc_id.vendor  = le16toh(UGETW(dd->idVendor));
  sc->sc_id.product = le16toh(UGETW(dd->idProduct));
  sc->sc_id.version = le16toh(UGETW(dd->bcdDevice));

  for (int i = 0; i < EVT_USED; ++i)
    if (!(sc->sc_bm[i] = malloc(bm_size[i], M_USBDEV, M_NOWAIT | M_ZERO)))
      goto nomem;

  if (!(sc->sc_btn = malloc(BTN_USED * sizeof(int), M_USBDEV, M_NOWAIT | M_ZERO)))
    goto nomem;

  if (!(sc->sc_abs = malloc(ABS_USED * sizeof(struct abs_axis), M_USBDEV, M_NOWAIT | M_ZERO)))
    goto nomem;

  BM_SET_BIT(sc->sc_bm[0], EVT_SYN);
  BM_SET_BIT(sc->sc_bm[0], EVT_BTN);
  BM_SET_BIT(sc->sc_bm[0], EVT_REL);
  BM_SET_BIT(sc->sc_bm[0], EVT_ABS);

  BM_SET_BIT(sc->sc_bm[1], BTN_LEFT);
  BM_SET_BIT(sc->sc_bm[1], BTN_RIGHT);
  BM_SET_BIT(sc->sc_bm[1], BTN_MIDDLE);
  BM_SET_BIT(sc->sc_bm[1], BTN_TOOL_PEN);
  BM_SET_BIT(sc->sc_bm[1], BTN_TOOL_ERASER);
  BM_SET_BIT(sc->sc_bm[1], BTN_TOOL_MOUSE);
  BM_SET_BIT(sc->sc_bm[1], BTN_STYLUS_1);
  BM_SET_BIT(sc->sc_bm[1], BTN_STYLUS_2);

  BM_SET_BIT(sc->sc_bm[2], REL_WHEEL);

  BM_SET_BIT(sc->sc_bm[3], ABS_X);
  BM_SET_BIT(sc->sc_bm[3], ABS_Y);
  BM_SET_BIT(sc->sc_bm[3], ABS_PRESSURE);
  BM_SET_BIT(sc->sc_bm[3], ABS_DISTANCE);
  BM_SET_BIT(sc->sc_bm[3], ABS_MISC);

  uwacom_init_abs(sc, ABS_X,        0, sc->sc_type->x_max,                                   4, 0);
  uwacom_init_abs(sc, ABS_Y,        0, sc->sc_type->y_max,                                   4, 0);
  uwacom_init_abs(sc, ABS_PRESSURE, 0, sc->sc_type->pressure_max,                            0, 0);
  uwacom_init_abs(sc, ABS_DISTANCE, 0, uwacom_protocols[sc->sc_type->protocol].distance_max, 0, 0);

  switch (sc->sc_type->protocol) {
    case CINTIQ:
    case INTUOS3L:
      uwacom_init_intuos3_large(sc);
    case INTUOS3S:
      uwacom_init_intuos(sc);
      break;

    case MYOFFICE:
      uwacom_init_myoffice(sc);
    case GRAPHIRE4:
      uwacom_init_graphire4(sc);
    case GRAPHIRE:
      ;
  }

  return 0;

nomem:
  printf("%s: out of memory\n", device_get_nameunit(sc->sc_dev));
  sc->sc_dying = 1;
  return 1;
}

static void uwacom_init_graphire4(struct uwacom_softc *sc) {
  BM_SET_BIT(sc->sc_bm[0], EVT_MSC);

  BM_SET_BIT(sc->sc_bm[1], BTN_MISC_0);
  BM_SET_BIT(sc->sc_bm[1], BTN_MISC_4);
  BM_SET_BIT(sc->sc_bm[1], BTN_TOOL_PAD);

  BM_SET_BIT(sc->sc_bm[4], MSC_SERIAL);

  sc->sc_tool[1] = BTN_TOOL_PAD;
  sc->sc_serial[1] = SERIAL_PAD_GRAPHIRE4;
}

static void uwacom_init_myoffice(struct uwacom_softc *sc) {
  BM_SET_BIT(sc->sc_bm[1], BTN_MISC_1);
  BM_SET_BIT(sc->sc_bm[1], BTN_MISC_5);

  BM_SET_BIT(sc->sc_bm[3], ABS_WHEEL);

  uwacom_init_abs(sc, ABS_WHEEL, 0, 71, 0, 0);
}

static void uwacom_init_intuos(struct uwacom_softc *sc) {
  BM_SET_BIT(sc->sc_bm[0], EVT_MSC);

  BM_SET_BIT(sc->sc_bm[1], BTN_MISC_0);
  BM_SET_BIT(sc->sc_bm[1], BTN_MISC_1);
  BM_SET_BIT(sc->sc_bm[1], BTN_MISC_2);
  BM_SET_BIT(sc->sc_bm[1], BTN_MISC_3);
  BM_SET_BIT(sc->sc_bm[1], BTN_SIDE);
  BM_SET_BIT(sc->sc_bm[1], BTN_EXTRA);
  BM_SET_BIT(sc->sc_bm[1], BTN_TOOL_PAD);

  BM_SET_BIT(sc->sc_bm[3], ABS_Z);
  BM_SET_BIT(sc->sc_bm[3], ABS_RX);
  BM_SET_BIT(sc->sc_bm[3], ABS_TILT_X);
  BM_SET_BIT(sc->sc_bm[3], ABS_TILT_Y);

  BM_SET_BIT(sc->sc_bm[4], MSC_SERIAL);

  uwacom_init_abs(sc, ABS_Z,      -900,  899, 0, 0);
  uwacom_init_abs(sc, ABS_RX,        0, 4096, 0, 0);
  uwacom_init_abs(sc, ABS_TILT_X,    0,  127, 0, 0);
  uwacom_init_abs(sc, ABS_TILT_Y,    0,  127, 0, 0);

  sc->sc_tool[1] = BTN_TOOL_PAD;
  sc->sc_tool_id[1] = TOOL_ID_PAD;
  sc->sc_serial[1] = SERIAL_PAD_INTUOS3;
}

static void uwacom_init_intuos3_large(struct uwacom_softc *sc) {
  BM_SET_BIT(sc->sc_bm[1], BTN_MISC_4);
  BM_SET_BIT(sc->sc_bm[1], BTN_MISC_5);
  BM_SET_BIT(sc->sc_bm[1], BTN_MISC_6);
  BM_SET_BIT(sc->sc_bm[1], BTN_MISC_7);

  BM_SET_BIT(sc->sc_bm[3], ABS_RY);

  uwacom_init_abs(sc, ABS_RY, 0, 4096, 0, 0);
}

static void uwacom_init_abs(struct uwacom_softc *sc, int axis, int32_t min, int32_t max, int32_t fuzz, int32_t flat) {
  sc->sc_abs[axis].min = min;
  sc->sc_abs[axis].max = max;
  sc->sc_abs[axis].fuzz = fuzz;
  sc->sc_abs[axis].flat = flat;
}

static void uwacom_clone(void *arg, struct ucred *cred, char *name, int namelen, struct cdev **dev) {
  if (*dev)
    return;

  int tablet;
  if (dev_stdclone(name, NULL, "event", &tablet) != 1)
    return;

  struct uwacom_softc *sc = devclass_get_softc(uwacom_devclass, tablet);
  if (!sc || sc->sc_dying)
    return;

  struct cdev *dev_handler;

  int max = -1;
  LIST_FOREACH(dev_handler, &sc->dev->si_children, si_siblings)
    if (!dev_handler->si_drv1) {
      *dev = dev_handler;
      return;
    } else if (HANDLER(dev_handler) > max)
      max = HANDLER(dev_handler);

  int handler;

  int *map = malloc((max + 1) * sizeof(int), M_USBDEV, M_NOWAIT | M_ZERO);
  if (map) {
    LIST_FOREACH(dev_handler, &sc->dev->si_children, si_siblings)
      map[HANDLER(dev_handler)] = 1;

    for (handler = 0; handler <= max; ++handler)
      if (!map[handler])
        break;

    free(map, M_USBDEV);
  } else
    handler = max + 1;

  if (handler > HANDLER_MAX)
    return;

  *dev = make_dev(&uwacom_cdevsw, MINOR(tablet, handler), UID_ROOT, GID_OPERATOR, 0644, "%s.%d", name, handler);
  dev_depends(sc->dev, *dev);
  (*dev)->si_flags |= SI_CHEAPCLONE;
  (*dev)->si_drv1 = NULL;
}

MODULE_DEPEND(uwacom, usb, 1, 1, 1);
DRIVER_MODULE(uwacom, uhub, uwacom_driver, uwacom_devclass, usbd_driver_load, NULL);
